# frozen_string_literal: true

require 'rails_helper'

describe SpiritHub::TagsetsController, type: :request do
  let(:current_user) { create(:user) }
  let(:tagset) { create(:tagset_doc, responsible: current_user.id, title: 'Doc Title 1') }

  before { sign_in(current_user) }

  describe 'POST #form' do
    subject { post "/spirit_hub/tagsets/form/#{state_key}", params: }

    let(:turbo_stream) do
      instance_double(Turbo::Streams::TagBuilder, update: '<turbo-stream />').tap do |mock|
        allow_any_instance_of(described_class).to receive(:turbo_stream).and_return(mock)
      end
    end

    before { SpiritHub::HubdocCycle.import_cycle_from_json }

    shared_examples 'validate correct partial rendering' do
      it 'renders the correct turbo modal' do
        expect(turbo_stream).to receive(:update).with(:modal, partial:, locals: hash_including(model: tagset))

        subject
      end
    end

    shared_examples 'validate correct user candidates for the dropdown' do
      let(:user_candidates) do
        ([current_user] + create_list(:user, 5)).map do |user|
          ["#{user.full_name} (#{user.email})", user.id.to_s]
        end
      end

      it 'renders the user candidates with a limit of 5' do
        expect(user_candidates.length).to eq 6

        expect(turbo_stream).to receive(:update).with(
          :modal, partial: anything, locals: hash_including(candidates_param_name => match_array(user_candidates[0..4]))
        )

        subject
      end

      context 'when there are candidate suggestions from the AI' do
        let(:user_candidates) do
          ([current_user] + create_list(:user, 10)).map do |user|
            ["#{user.full_name} (#{user.email})", user.id.to_s]
          end
        end

        let(:ai_suggestion_ids) do
          id = 1 # index

          [
            user_candidates[2][id],
            user_candidates[8][id],
            user_candidates[9][id]
          ]
        end

        it 'pre-loads all the suggestions as candidates, even if it passes the 5 limit' do
          allow_any_instance_of(SpiritHub::TagsetSuggester).to receive(suggester_method).and_return(ai_suggestion_ids)

          expected_candidates = [
            user_candidates[0],
            user_candidates[1],
            user_candidates[2],
            user_candidates[3],
            user_candidates[4], # <--- limit of 5
            user_candidates[8],
            user_candidates[9]
          ]

          expect(turbo_stream).to receive(:update).with(
            :modal, partial: anything, locals: hash_including(candidates_param_name => match_array(expected_candidates), recipients_suggestions: ai_suggestion_ids)
          )

          subject
        end
      end
    end

    context 'when :share form' do
      let(:state_key) { 'share' }
      let(:params) { { ids: tagset.id.to_s } }

      it 'returns 200' do
        subject
        expect(response).to have_http_status(200)
      end

      include_examples 'validate correct partial rendering' do
        let(:partial) { 'spirit_hub/common/state_forms/share_form' }
      end

      include_examples 'validate correct user candidates for the dropdown' do
        let(:candidates_param_name) { :share_candidates }
        let(:suggester_method) { :suggest_recipients_to_share_with }
      end
    end

    context 'when :to_answer form' do
      let(:state_key) { 'to_answer' }
      let(:params) { { ids: tagset.id.to_s } }

      it 'returns 200' do
        subject
        expect(response).to have_http_status(200)
      end

      include_examples 'validate correct partial rendering' do
        let(:partial) { 'spirit_hub/common/state_forms/send_to_form' }
      end

      include_examples 'validate correct user candidates for the dropdown' do
        let(:candidates_param_name) { :send_to_user_groups }
        let(:suggester_method) { :suggest_recipients_to_answer }
      end
    end

    context 'when :to_sign form' do
      let(:state_key) { 'to_sign' }
      let(:params) { { ids: tagset.id.to_s } }

      it 'returns 200' do
        subject
        expect(response).to have_http_status(200)
      end

      include_examples 'validate correct partial rendering' do
        let(:partial) { 'spirit_hub/common/state_forms/to_sign_form' }
      end

      include_examples 'validate correct user candidates for the dropdown' do
        let(:candidates_param_name) { :user_candidates }
        let(:suggester_method) { :suggest_recipients_to_sign }
      end
    end
  end

  describe 'POST #versioning' do
    subject { post "/spirit_hub/tagsets/#{tagset.id}/versioning", params: @params }

    let(:doc) { create :doc }

    before do
      SpiritHub::HubdocCycle.import_cycle_from_json
      tagset.update(doc_ids: [doc.id])
    end

    context 'when new uploaded doc' do
      before do
        file_path = Rails.root.join('packages/spirit_hub/spec/fixtures/files/one_page.pdf')
        file = Rack::Test::UploadedFile.new(file_path, 'application/pdf')
        @params = { doc_id: tagset.doc.id.to_s }.merge(new_version_file: file)
      end

      it 'succeeds' do
        subject
        expect(response).to have_http_status 200
        expect(flash[:notice]).to eq 'Nouvelle version du document ajoutée.'
      end

      it 'creates a new doc as version' do
        expect { subject }.to change { tagset.reload.all_docs.count }
          .by(1)
          .and change { tagset.reload.title }.from('Doc Title 1').to('one_page.pdf')
        expect(tagset.doc.versions_count).to eq 1
      end
    end

    context 'when existing doc' do
      before do
        @another_tagset = create(:tagset_doc, :with_file, filename: 'pdf_sample.pdf', responsible: current_user.id)
        @params = { doc_id: tagset.doc.id.to_s }.merge(existing_doc_id: @another_tagset.doc.id.to_s)
      end

      it 'succeeds' do
        subject
        expect(response).to have_http_status 200
        expect(flash[:notice]).to eq 'Nouvelle version du document ajoutée.'
      end

      it 'updates tagset with existing doc as version' do
        expect { subject }.to change { tagset.reload.all_docs.count }
          .by(1)
          .and change { tagset.reload.title }.from('Doc Title 1').to('pdf_sample.pdf')
        expect(tagset.doc.versions_count).to eq 1
      end
    end
  end

  describe 'POST #detach_version' do
    subject { post "/spirit_hub/tagsets/#{tagset.id}/detach_version/#{version.id}" }

    let(:doc) { create :doc }
    let(:version) { create(:doc, version_of: doc.id.to_s) }

    before do
      SpiritHub::HubdocCycle.import_cycle_from_json
      tagset.update(doc_ids: [doc.id, version.id])
    end

    it 'detaches the version' do
      expect { subject }.to change { version.reload.version_of }
        .from(doc.id).to(nil)
        .and change { tagset.reload.docs.count }.from(1).to(2)
      expect(doc.versions_count).to eq 0
    end
  end

  describe 'GET #share_from_info_form' do
    subject { get "/spirit_hub/tagsets/#{tagset.id}/share_from_info_form" }

    let(:turbo_stream) do
      instance_double(Turbo::Streams::TagBuilder, update: '<turbo-stream />').tap do |mock|
        allow_any_instance_of(described_class).to receive(:turbo_stream).and_return(mock)
      end
    end

    it 'renders the correct turbo modal' do
      expect(turbo_stream).to receive(:update).with(
        :modal,
        partial: 'spirit_hub/tagsets/share_from_info_form',
        locals: hash_including(model: tagset)
      )

      subject
    end
  end

  describe 'POST #share_from_info' do
    subject { post "/spirit_hub/tagsets/#{tagset.id}/share_from_info", params: }

    let(:params) do
      {
        'subject' => 'Hey',
        'message' => 'Hey there',
        'docs' => {
          docs[0].id.to_s => { 'to' => [users[0].id.to_s, users[1].id.to_s] },
          docs[1].id.to_s => { 'to' => [users[2].id.to_s] }
        }
      }
    end

    let(:users) { create_list(:user, 3) }
    let(:docs) { create_list(:doc, 2) }

    it 'calls the service to share the docs accordingly' do
      allow(SpiritHub::TagsetShare).to receive(:call)

      subject

      expect(SpiritHub::TagsetShare).to have_received(:call).with(
        tagset:,
        doc_ids: [docs[0].id.to_s],
        to: match_array([users[0].id.to_s, users[1].id.to_s]),
        subject: 'Hey',
        message: 'Hey there',
        current_user:
      ).once

      expect(SpiritHub::TagsetShare).to have_received(:call).with(
        tagset:,
        doc_ids: [docs[1].id.to_s],
        to: match_array([users[2].id.to_s]),
        subject: 'Hey',
        message: 'Hey there',
        current_user:
      ).once
    end
  end
end
